import sys

from rich import print

from unpage.cli.mcp.tools._app import tools_app
from unpage.config import manager
from unpage.knowledge import Graph
from unpage.mcp import Context, build_mcp_server
from unpage.plugins import PluginManager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@tools_app.command(name="list")
async def list_tools() -> None:
    """List all MCP tools available from enabled plugins."""
    await telemetry.send_event(
        {
            "command": "mcp tools list",
            **prepare_profile_for_telemetry(manager.get_active_profile()),
        }
    )
    config = manager.get_active_profile_config()
    plugins = PluginManager(config=config)
    context = Context(
        profile=manager.get_active_profile(),
        config=config,
        plugins=plugins,
        graph=Graph(manager.get_active_profile_directory() / "graph.json"),
    )
    mcp = await build_mcp_server(context)

    tools = await mcp.get_tools()
    if not tools:
        print("[red]No MCP tools available from enabled plugins.[/red]")
        print("[bold]Enable plugins with 'unpage configure' to access more tools.[/bold]")
        sys.exit(1)

    for key, tool in tools.items():
        cmd = [key]
        for arg, arg_data in tool.parameters["properties"].items():
            arg_type = arg_data.get("type", "unknown")
            if "anyOf" in arg_data:
                arg_type = "|".join(t["type"] for t in arg_data["anyOf"])
            cmd.append(f"<{arg}:{arg_type}>")
        print(" ".join(cmd))



import rich

from unpage import __version__
from unpage.cli._app import app
from unpage.telemetry import client as telemetry


@app.command
async def version(*, json: bool = False) -> None:
    """Display the version of the Unpage CLI.

    Parameters
    ----------
    json
        Return the version information as JSON
    """
    await telemetry.send_event(
        {
            "command": "version",
            "json_output": json,
        }
    )

    if json:
        rich.print_json(
            data={
                "unpage": __version__,
            }
        )
        return
    print(f"unpage {__version__}")



import re
from typing import Any

import networkx as nx

from unpage.knowledge import Graph
from unpage.plugins.base import Plugin
from unpage.plugins.mixins import McpServerMixin, tool
from unpage.utils import compile_regex


class GraphPlugin(Plugin, McpServerMixin):
    @property
    def graph(self) -> Graph:
        return self.context.graph

    @tool()
    async def search_resources(self, identifier_or_regex: str) -> list[str] | str:
        """Find resources with an identifier matching the given identifier or
        regular expression.

        To use a regular expression, surround it with slashes. For example,
        `/^my-resource-\\d+$/` will match any resource that starts with
        `my-resource-` and is followed by one or more digits.

        Returns the node IDs, which can be used with get_resource_details to
        get the full resource.
        """
        is_regex = re.match(r"^/.*/[gimsxu]*$", identifier_or_regex) is not None

        # Compile the pattern if it's a regex (or turn the string into an exact-match regex)
        pattern = compile_regex(
            identifier_or_regex if is_regex else rf"/^{re.escape(identifier_or_regex)}$/"
        )

        results = []
        async for node in self.graph.iter_nodes():
            for identifier in (node.nid, *await node.get_identifiers()):
                if not identifier:
                    continue
                if pattern.match(identifier):
                    results.append(node.nid)

        # If there were no results with an exact match, try to fuzzy match any part of identifiers
        if not results and not is_regex:
            fuzzy_pattern = compile_regex(rf"/.*{re.escape(identifier_or_regex)}.*/")
            async for node in self.graph.iter_nodes():
                for identifier in (node.nid, *await node.get_identifiers()):
                    if not identifier:
                        continue
                    if fuzzy_pattern.match(identifier):
                        results.append(node.nid)

        # If there were a lot of results, truncate to 100 and add a message
        if len(results) > 500:
            results = results[:500]
            results.append(
                "Note: There were too many results to return all of them. Consider using other tools to gain a better understanding of the system and refining your search."
            )

        return results or "No resources found matching that identifier."

    @tool()
    async def get_resource_details(self, node_id: str) -> dict[str, Any] | str:
        """Get the full details of a resource from its node ID."""
        node = await self.graph.get_node_safe(node_id)
        if not node:
            return f"Resource with node ID '{node_id}' not found"
        return node.raw_data

    @tool()
    async def get_resource_topology(self) -> str:
        """Get a map of the types of resources and how they're connected.

        You should call this first before trying to search for resources.
        """
        topology = nx.DiGraph()

        async for edge in self.graph.iter_edges():
            source_type = edge.source_node.node_type
            destination_type = edge.destination_node.node_type

            # Add nodes (types) and edge between them
            topology.add_edge(
                source_type,
                destination_type,
                label=edge.properties["relationship_type"],
            )

        return str(nx.nx_pydot.to_pydot(topology))

    @tool()
    async def get_resource_map(
        self,
        root_node_id: str,
        max_depth: int = 2,
    ) -> str:
        """Get a detailed map of a resource and its dependencies.

        This is a good first tool to use when you want to get a high-level
        understanding of some part of the system.

        The root node is the starting point of the map. This tool will return
        all resources directly related to the root resource, and all resources
        related to those resources (etc), up to the given max depth.
        """
        resource_map = nx.DiGraph()

        node = await self.graph.get_node_safe(root_node_id)
        if not node:
            return f"Resource with node ID '{root_node_id}' not found"

        async for edge in self.graph.iter_neighborhood_edges(node, max_depth):
            resource_map.add_edge(
                f'"{edge.source_node.nid}"',
                f'"{edge.destination_node.nid}"',
                label=edge.properties["relationship_type"],
            )

        return str(nx.nx_pydot.to_pydot(resource_map))

    @tool()
    async def get_neighboring_resources(self, node_id: str, max_depth: int = 1) -> list[str] | str:
        """Get the IDs of the immediate neighbors of the resource with the given node ID.

        Use get_resource_details to get the full details of a resource.
        """
        node = await self.graph.get_node_safe(node_id)
        if not node:
            return f"Resource with node ID '{node_id}' not found"
        return [n.nid async for n in self.graph.iter_neighbors(node)]



import asyncio
import os
import sys
from collections.abc import Awaitable, Callable
from datetime import UTC, datetime
from typing import Any, cast

import human_readable
import questionary
import rich
from pydantic import BaseModel
from questionary import Choice
from rich.console import Console
from rich.panel import Panel

from unpage.agent.analysis import AnalysisAgent
from unpage.agent.utils import load_agent
from unpage.cli.agent._app import agent_app
from unpage.cli.agent.create import create_agent
from unpage.cli.configure import welcome_to_unpage
from unpage.config import Config, PluginConfig, manager
from unpage.plugins.base import PluginManager
from unpage.plugins.datadog.plugin import DatadogPlugin
from unpage.plugins.llm.plugin import LlmPlugin
from unpage.plugins.pagerduty.models import PagerDutyIncident
from unpage.plugins.pagerduty.plugin import PagerDutyPlugin
from unpage.plugins.solarwinds.plugin import SolarWindsPlugin
from unpage.telemetry import client as telemetry
from unpage.telemetry import hash_value, prepare_profile_for_telemetry
from unpage.utils import confirm, edit_file, select


async def _send_event(step: str, extra_params: dict[Any, Any] | None = None) -> None:
    await telemetry.send_event(
        {
            "command": "agent quickstart",
            "step": step,
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            **(extra_params if extra_params else {}),
        }
    )


@agent_app.command
async def quickstart() -> None:
    """Get up-and-running with an incident agent in less than 5 minutes!"""
    await _send_event("start")
    welcome_to_unpage()
    _quickstart_intro()
    config = manager.get_empty_config(
        manager.get_active_profile(), plugins=_initial_plugin_settings()
    )
    cfg, next_step_count = await _create_config(config)
    plugin_manager = PluginManager(cfg)
    cfg.save()
    await _send_event("config_saved")
    agent_name = await _create_and_edit_agent(next_step_count)
    await _send_event("agent_created")
    await _demo_an_incident(agent_name, next_step_count + 1, plugin_manager)
    await _send_event("incident_demoed")
    await _show_agent_commands(next_step_count + 2)
    await _send_event("shown_agent_commands")
    await _optionally_launch_configure(next_step_count + 3)


def _quickstart_intro() -> None:
    rich.print("""This interactive tool will walk through the configuration of your system.
We'll setup basic plugins so your LLM can use the Unpage MCP server.


Here's what the quickstart will entail:

1. Configure LLM (Amazon Bedrock, OpenAI, Anthropic Claude, and many more!)
2. Configure Pagerduty plugin
3. Optionally configure SolarWinds
4. Optionally configure Datadog
5. Create and edit a demo agent
6. Test the demo agent against a PagerDuty incident
7. See how to create, test and refine your own agents
8. Configure and build an infrastructure knowledge graph
""")


def _initial_plugin_settings() -> dict[str, PluginConfig]:
    try:
        existing_config = manager.get_active_profile_config()
    except FileNotFoundError:
        existing_config = manager.get_empty_config(manager.get_active_profile())

    return {
        "core": PluginConfig(enabled=True),
        "networking": PluginConfig(enabled=True),
        "llm": PluginConfig(
            enabled=True,
            settings=(
                LlmPlugin.default_plugin_settings
                if "llm" not in existing_config.plugins
                else existing_config.plugins["llm"].settings
            ),
        ),
        "pagerduty": PluginConfig(
            enabled=True,
            settings=(
                PagerDutyPlugin.default_plugin_settings
                if "pagerduty" not in existing_config.plugins
                else existing_config.plugins["pagerduty"].settings
            ),
        ),
        "solarwinds": PluginConfig(
            enabled=False,
            settings=(
                SolarWindsPlugin.default_plugin_settings
                if "solarwinds" not in existing_config.plugins
                else existing_config.plugins["solarwinds"].settings
            ),
        ),
        "datadog": PluginConfig(
            enabled=False,
            settings=(
                DatadogPlugin.default_plugin_settings
                if "datadog" not in existing_config.plugins
                else existing_config.plugins["datadog"].settings
            ),
        ),
    }


async def _create_config(cfg: Config) -> tuple[Config, int]:
    plugin_manager = PluginManager(cfg)
    required_plugins = [
        "llm",
        "pagerduty",
    ]
    rich.print(
        "Next we're going to configure plugins! Plugins are vendor specific integrations to Unpage"
    )
    rich.print("")
    await questionary.press_any_key_to_continue().unsafe_ask_async()
    rich.print("")
    for i, plugin in enumerate(required_plugins):
        console = Console()
        console.print(
            Panel(
                f"[bold]{i + 1}. {plugin.upper() if plugin == 'llm' else plugin.capitalize()} configuration[/bold]",
                width=80,
            )
        )
        attempts = 1
        while True:
            cfg.plugins[plugin].settings = await plugin_manager.get_plugin(
                plugin
            ).interactive_configure()
            plugin_manager = PluginManager(cfg)
            if await _plugin_valid(plugin_manager, plugin):
                await _send_event(
                    f"plugin_valid_{plugin}",
                    extra_params={
                        "attempts": attempts,
                    },
                )
                break
            rich.print(f"Validation failed for {plugin}")
            if not await confirm("Retry?"):
                await _send_event(
                    f"plugin_invalid_{plugin}",
                    extra_params={
                        "attempts": attempts,
                    },
                )
                break
            attempts += 1
            rich.print("")
        rich.print("")
    optional_plugins = [
        "solarwinds",
        "datadog",
    ]
    for i, optional_plugin in enumerate(optional_plugins):
        console = Console()
        console.print(
            Panel(
                f"[bold]{i + len(required_plugins) + 1}. {optional_plugin.upper() if optional_plugin == 'llm' else optional_plugin.capitalize()} configuration (optional)[/bold]",
                width=80,
            )
        )
        if await confirm(
            f"Would you like to enable and configure {optional_plugin.upper() if optional_plugin == 'llm' else optional_plugin.capitalize()}",
            default=False,
        ):
            cfg.plugins[optional_plugin].enabled = True
            attempts = 1
            while True:
                cfg.plugins[optional_plugin].settings = await plugin_manager.get_plugin(
                    optional_plugin
                ).interactive_configure()
                plugin_manager = PluginManager(cfg)
                if await _plugin_valid(plugin_manager, optional_plugin):
                    await _send_event(
                        f"plugin_valid_{optional_plugin}",
                        extra_params={
                            "attempts": attempts,
                        },
                    )
                    break
                rich.print(f"Validation failed for {optional_plugin}")
                if not await confirm("Retry?"):
                    await _send_event(
                        f"plugin_invalid_{optional_plugin}",
                        extra_params={
                            "attempts": attempts,
                        },
                    )
                    break
                attempts += 1
        else:
            await _send_event(f"plugin_disabled_{optional_plugin}")
        rich.print("")
    next_step_count = len(required_plugins) + len(optional_plugins) + 1
    return (cfg, next_step_count)


async def _plugin_valid(plugin_manager: PluginManager, plugin: str) -> bool:
    rich.print(f"Validating {plugin}...")
    try:
        await plugin_manager.get_plugin(plugin).validate_plugin_config()
    except Exception as ex:
        rich.print(f"Error validating {plugin}:\n{ex}")
        return False
    rich.print(f"[green]{plugin} configuration is valid![/green]")
    return True


async def _create_and_edit_agent(next_step_count: int) -> str:
    console = Console()
    console.print(Panel(f"[bold]{next_step_count}. Create and edit demo agent[/bold]", width=80))
    agent_name = "demo-quickstart"
    template = "demo_quickstart"
    agent_file = await create_agent(agent_name, True, template)
    await _send_event(
        "created_agent",
        extra_params={
            "agent_name_sha256": hash_value(agent_name),
            "template": template,
        },
    )
    rich.print("")
    rich.print(f"> We created a new agent called {agent_name}!")
    rich.print(
        "> When you're ready, we'll open the agent template in your editor so you can preview it and optionally make changes before testing."
    )
    rich.print("")
    await questionary.press_any_key_to_continue().unsafe_ask_async()
    await edit_file(agent_file)
    rich.print("")
    return agent_name


async def _enter_incident_id_or_url(pd: PagerDutyPlugin) -> PagerDutyIncident | None:
    while True:
        answer = await questionary.text(
            "PagerDuty incident ID or URL",
        ).unsafe_ask_async()
        incident_id = answer
        if "/" in answer:
            incident_id = [x for x in answer.split("/") if x][-1]
        try:
            return await pd.get_incident_by_id(incident_id)
        except Exception as ex:
            rich.print(f"Failed to retrieve incident with id {incident_id}: {ex}")
            if not await confirm("Retry with another id or url?"):
                return None


async def _select_incident_from_recent_100(pd: PagerDutyPlugin) -> PagerDutyIncident | None:
    incidents: list[PagerDutyIncident] = []
    console = Console()
    with console.status("querying incidents...", spinner="dots") as status:
        async for incident in pd.recent_incident_payloads():
            incidents.append(incident.incident)
            if len(incidents) >= 100:
                break
        status.update("Done 🎉")
    enable_search = len(incidents) > 10
    incident_id = await select(
        "Select a PagerDuty incident",
        choices=[
            Choice(
                f"{i.title} [{i.urgency}] [{i.id}]",
                value=i.id,
            )
            for i in incidents
        ],
        use_search_filter=enable_search,
        use_jk_keys=not enable_search,
    )
    for incident in incidents:
        if incident.id == incident_id:
            return incident


async def _random_incident_from_recent(pd: PagerDutyPlugin) -> PagerDutyIncident | None:
    async for incident in pd.recent_incident_payloads():
        return incident.incident


async def _select_pagerduty_incident(pd: PagerDutyPlugin) -> PagerDutyIncident | None:
    class incidentChooser(BaseModel):
        title: str
        func: Callable[[PagerDutyPlugin], Awaitable[PagerDutyIncident | None]]

    opts = [
        incidentChooser(
            title="Enter an incident id or url",
            func=_enter_incident_id_or_url,
        ),
        incidentChooser(
            title="Select from the most recent 100 incidents",
            func=_select_incident_from_recent_100,
        ),
        incidentChooser(
            title="Have us select a recent incident",
            func=_random_incident_from_recent,
        ),
    ]
    while True:
        choice = await select(
            "How would you like to select a PagerDuty incident?",
            choices=[Choice(o.title, value=str(i)) for i, o in enumerate(opts)],
        )
        incident = await opts[int(choice)].func(pd)
        if incident:
            await _send_event(
                "selected_pagerduty_incident",
                extra_params={"selection_method": opts[int(choice)].title},
            )
            return incident
        rich.print("Oops, did not get an incident id to test with")
        if not await confirm("Retry?"):
            return None


async def _demo_an_incident(
    agent_name: str, next_step_count: int, plugin_manager: PluginManager
) -> None:
    console = Console()
    console.print(Panel(f"[bold]{next_step_count}. Demo time![/bold]", width=80))
    rich.print("")
    rich.print("Now we can test the agent output with one of your existing PagerDuty incidents!")
    rich.print("")
    pd = cast("PagerDutyPlugin", plugin_manager.get_plugin("pagerduty"))
    incident = await _select_pagerduty_incident(pd)
    if not incident:
        rich.print("Did not get an incident, skipping the demo.")
        return
    try:
        agent = load_agent(agent_name)
        analysis_agent = AnalysisAgent()
        rich.print("")
        rich.print("Details of the incident we're going to demo:")
        rich.print(f"> Title: {incident.title}")
        time_since = datetime.now(UTC) - incident.created_at
        rich.print(
            f"> Created: {incident.created_at} ({human_readable.precise_delta(time_since)} ago)"
        )

        def _color(status: str) -> str:
            match status:
                case "triggered":
                    return f"[red]{incident.status}[/red]"
                case "acknowledged":
                    return f"[yellow]{incident.status}[/yellow]"
            return incident.status

        rich.print(f"> Status: {_color(incident.status)}")
        rich.print(f"> Urgency: {incident.urgency}")
        rich.print(f"> Url: {incident.html_url}")
        rich.print("")
        rich.print("> Ready to run the demo agent on this incident?")
        rich.print("")
        await questionary.press_any_key_to_continue().unsafe_ask_async()
        incident_json = incident.model_dump_json(indent=2)
        incident_json_lines = incident_json.splitlines()
        truncated = False
        if len(incident_json_lines) > 20:
            incident_json = "\n".join(incident_json_lines[-20:])
            truncated = True
        rich.print("")
        rich.print(f"> PagerDuty incident payload{' (last 20 lines)' if truncated else ''}:")
        if truncated:
            rich.print("...")
        rich.print(incident_json)
        rich.print("> Computing status update... (this may take a minute!)")
        console = Console()
        with console.status("working...", spinner="dots") as status:
            result = await analysis_agent.acall(payload=incident.model_dump_json(), agent=agent)
            status.update("Done 🎉")
        rich.print("")
        rich.print(
            f"> Status update that would be posted to PagerDuty by the {pd.default_from} user:\n"
        )
        rich.print("-----")
        rich.print(result)
        rich.print("-----")
        rich.print("")
        rich.print("You can re-run this demo at any point with:")
        rich.print("")
        rich.print(
            f"  [bold deep_sky_blue1]uvx unpage agent run --pagerduty-incident {incident.id} demo-quickstart[/bold deep_sky_blue1]"
        )
        rich.print("")
    except Exception as ex:
        rich.print(f"[red] Demo failed:[/red] {ex}")
        sys.exit(1)
    rich.print("> Congrats on completing the demo agent! You did it! 🎉")
    rich.print("")
    rich.print("> Ready to move on?")
    rich.print("")
    await questionary.press_any_key_to_continue().unsafe_ask_async()
    rich.print("")


async def _show_agent_commands(next_step_count: int) -> None:
    console = Console()
    console.print(
        Panel(f"[bold]{next_step_count}. Create, test and refine your own agents[/bold]", width=80)
    )
    rich.print("")
    rich.print(
        "> You can create, edit, run, and serve agents using the [bold deep_sky_blue1]uvx unpage agent[/bold deep_sky_blue1] subcommands:"
    )
    rich.print("> ")
    agent_help_cmd = " ".join([a if a != "quickstart" else "--help" for a in sys.argv])
    rich.print("> $ [bold deep_sky_blue1]uvx unpage agent --help[/bold deep_sky_blue1]")
    rich.print("")
    await (await asyncio.subprocess.create_subprocess_shell(agent_help_cmd)).wait()
    rich.print("> ")
    rich.print("")
    rich.print(
        "> Next, you can build the knowledge graph of your infrastructure. This will give your agents more tools and context, so you get better results. Ready to learn more about the knowledge graph?"
    )
    rich.print("")
    await questionary.press_any_key_to_continue().unsafe_ask_async()
    rich.print("")


async def _optionally_launch_configure(next_step_count: int) -> None:
    console = Console()
    console.print(
        Panel(f"[bold]{next_step_count}. Infrastructure Knowledge Graph[/bold]", width=80)
    )
    rich.print("")
    rich.print(
        "> Unpage supports a rich infrastructure knowledge graph builder, which can be used on its own or automatically "
        "integrated with your Unpage Agents. The graph can be built from your infrastructure tools, like AWS or Aptible, and your "
        "observability tools, like Datadog and CloudWatch."
    )
    rich.print("")
    rich.print(
        "> Use the [bold deep_sky_blue1]uvx unpage configure[/bold deep_sky_blue1] command to configure all plugins required for graph building."
    )
    rich.print(
        "> Then use [bold deep_sky_blue1]uvx unpage graph build[/bold deep_sky_blue1] to build the infrastructure knowledge graph."
    )
    rich.print(">")
    rich.print(
        "> Your Unpage Agents will automatically begin using the knowledge graph once it is built."
    )
    rich.print(">")
    if not await confirm("Would you like to run uvx unpage configure now?"):
        await _send_event("done_no_configure")
        return
    rich.print(">")
    await _send_event("starting_configure")
    _replace_current_proc_with_unpage_configure()


def _replace_current_proc_with_unpage_configure() -> None:
    configure_cmd = [a if a != "agent" else "configure" for a in sys.argv if a != "quickstart"]
    rich.print("> Running: [bold deep_sky_blue1]uvx unpage configure[/bold deep_sky_blue1]")
    rich.print("")
    os.execvp(configure_cmd[0], configure_cmd)  # noqa: S606 Starting a process without a shell



import sys

import rich

from unpage.cli.profile._app import profile_app
from unpage.config import manager


@profile_app.command
def create(profile_name: str) -> None:
    """Create a new profile.

    Parameters
    ----------
    profile_name
        The name of the profile to create
    """
    try:
        manager.create_profile(profile_name)
        rich.print(f"[green]Profile '{profile_name}' created successfully![/green]")
    except FileExistsError:
        rich.print(f"[red]Profile '{profile_name}' already exists![/red]")
        sys.exit(1)



import sys

from rich import print

from unpage.cli.agent._app import agent_app
from unpage.cli.agent.actions import create_agent
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import hash_value, prepare_profile_for_telemetry
from unpage.utils import edit_file


@agent_app.command
async def create(
    agent_name: str,
    /,
    *,
    overwrite: bool = False,
    template: str = "default",
    editor: str | None = None,
    no_edit: bool = False,
) -> None:
    """Create a new agent configuration file and open it in your editor.

    Parameters
    ----------
    agent_name
        The name of the agent to create
    overwrite
        Overwrite the agent file if it already exists
    template
        The template to use to create the agent file
    editor
        The editor to use to open the agent file
    no_edit
        Do not open the agent file in your editor
    """
    await telemetry.send_event(
        {
            "command": "agent create",
            "agent_name_sha256": hash_value(agent_name),
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "overwrite": overwrite,
            "template": template,
            "editor": editor,
            "no_edit": no_edit,
        }
    )
    agent_file = await create_agent(
        agent_name=agent_name,
        overwrite=overwrite,
        template=template,
    )

    # Skip file editing if specified.
    if no_edit:
        return

    # Open the file in the user's editor
    try:
        await edit_file(agent_file, editor)
    except ValueError:
        print(
            "[red]No editor specified. Set the $EDITOR environment variable or pass the --editor option. Please manually open the file in your editor.[/red]",
            file=sys.stderr,
        )
        sys.exit(1)



from cyclopts import App

from unpage.cli.mcp.client.logs.app import client_logs_app

client_app = App(help="Debugging tools for clients of the Unpage MCP Server")
client_app.command(client_logs_app, name="logs")



import os
from collections.abc import AsyncGenerator
from typing import Any, cast

import questionary
from pagerduty.rest_api_v2_client import RestApiV2Client
from pydantic import AwareDatetime, BaseModel

from unpage.config import PluginSettings
from unpage.plugins.base import Plugin
from unpage.plugins.mixins import McpServerMixin, tool
from unpage.plugins.pagerduty.models import PagerDutyIncident, PagerDutyIncidentPayload
from unpage.utils import classproperty


class PagerDutyPluginSettings(BaseModel):
    api_key: str = ""
    default_from: str = ""


class PagerDutyPlugin(Plugin, McpServerMixin):
    def __init__(
        self,
        *args: Any,
        api_key: str | None = None,
        default_from: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._api_key = api_key or os.getenv("PAGERDUTY_API_KEY", "")
        self.default_from = default_from or os.getenv("PAGERDUTY_DEFAULT_FROM")
        if self._api_key:
            self._client = RestApiV2Client(self._api_key, default_from=self.default_from)

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return PagerDutyPluginSettings(
            api_key=os.environ.get("PAGERDUTY_API_KEY", ""),
            default_from=os.environ.get("PAGERDUTY_DEFAULT_FROM", ""),
        ).model_dump()

    async def interactive_configure(self) -> PluginSettings:
        defaults = self.default_plugin_settings
        return PagerDutyPluginSettings(
            api_key=await questionary.password(
                "API Key",
                default=self._api_key or defaults.get("api_key", ""),
                instruction="Generate a token with https://docs.unpage.ai/plugins/pagerduty#prerequisites",
            ).unsafe_ask_async(),
            default_from=await questionary.text(
                "Default User Email",
                default=self.default_from or defaults.get("default_user_email", ""),
                instruction="The email you provide will appear as the author for annotations left by the agent.",
            ).unsafe_ask_async(),
        ).model_dump()

    async def validate_plugin_config(self) -> None:
        if not self._api_key:
            raise LookupError(
                "plugins.pagerduty.settings.api_key (or $PAGERDUTY_API_KEY) must be set"
            )
        if not self.default_from:
            raise ValueError(
                "PagerDuty default_user_email may not be left empty or blank. Choose a user that will be used as the author when creating status updates."
            )
        for user in self._client.iter_all("users"):
            if user["email"].lower() == self.default_from.lower():
                return
        raise ValueError(f"No PagerDuty users have the email '{self.default_from}'")

    @tool()
    async def get_incident_details(self, incident_id: str) -> dict[str, Any]:
        """Get a PagerDuty incident by ID, including all alert details.

        Args:
            incident_id (str): The ID of the PagerDuty incident. Typically a
            string of uppercase letters and numbers. For example "PGR0VU2",
            "PF9KMXH", or "Q2K78SNJ5U1VE1".

        Returns:
            dict: The incident JSON payload, including all alert details.
        """
        incident = cast("dict[str, Any]", self._client.rget(f"/incidents/{incident_id}"))
        incident["alerts"] = await self.get_alert_details_for_incident(incident_id)
        return incident

    @tool()
    async def get_alert_details_for_incident(self, incident_id: str) -> list[dict[str, Any]]:
        """Get the details of the alert(s) for a PagerDuty incident.

        Args:
            incident_id (str): The ID of the PagerDuty incident. Typically a
            string of uppercase letters and numbers. For example "PGR0VU2",
            "PF9KMXH", or "Q2K78SNJ5U1VE1".

        Returns:
            list[dict]: The list of alert details.
        """
        return cast("list[dict[str, Any]]", self._client.rget(f"/incidents/{incident_id}/alerts"))

    @tool()
    async def post_status_update(self, incident_id: str, message: str) -> None:
        """Post a status update to a PagerDuty incident

        Args:
            incident_id (str): The PagerDuty ID of the incident
            message (str): The message to post
        """
        self._client.rpost(
            f"/incidents/{incident_id}/status_updates",
            json={
                "message": message,
            },
        )

    @tool()
    async def resolve_incident(
        self, incident_id: str, resolution_message: str | None = None
    ) -> None:
        """Resolve a PagerDuty incident

        Args:
            incident_id (str): The ID of the PagerDuty incident to resolve
            resolution_message (str, optional): A message to include with the resolution
        """
        data = {"incident": {"type": "incident", "status": "resolved"}}

        self._client.rput(
            f"/incidents/{incident_id}",
            json=data,
        )

        # If a resolution message was provided, add it as a status update
        if resolution_message:
            await self.post_status_update(incident_id, resolution_message)

    async def get_incident_by_id(self, incident_id: str) -> PagerDutyIncident:
        """Get a single incident by its id

        Args:
            incident_id (str): The PagerDuty ID of the incident

        Returns:
            PagerDutyIncident: the pagerduty incident data
        """
        return PagerDutyIncident(
            **cast("dict[str, Any]", self._client.rget(f"/incidents/{incident_id}"))
        )

    async def recent_incident_payloads(
        self, since: AwareDatetime | None = None, sort_by: str = "incident_number:desc"
    ) -> AsyncGenerator[PagerDutyIncidentPayload, None]:
        """Get a list of recent PagerDuty incidents.

        Returns:
            list: incident JSON payloads that can be used for demoing and testing agents
        """
        for _, incident in enumerate(
            self._client.iter_all(
                "incidents",
                params={
                    "statuses": ["triggered", "acknowledged", "resolved"],
                    "sort_by": sort_by,
                    **({"since": since.date().isoformat()} if since else {}),
                },
            )
        ):
            yield PagerDutyIncidentPayload(incident=PagerDutyIncident(**incident))



import base64
import hashlib
import json
import os
import sys
import uuid
from importlib.metadata import version
from typing import Any

import httpx
import rich
import sentry_sdk

from unpage.config import manager


def _get_or_create_user_id() -> str:
    identity_file = manager.config_root / ".identity"
    if not identity_file.exists():
        if not identity_file.parent.exists():
            identity_file.parent.mkdir(parents=True)
        identity_file.write_text(str(uuid.uuid4()))
    return identity_file.read_text().strip()


def hash_value(value: str) -> str:
    hasher = hashlib.sha256()
    hasher.update(value.encode("utf-8"))
    return base64.urlsafe_b64encode(hasher.digest()).decode("utf-8")


def prepare_profile_for_telemetry(profile: str) -> dict[str, Any]:
    return {
        "profile_sha256": hash_value(profile),
        **({"profile": "default"} if profile == "default" else {}),
    }


UNPAGE_TELEMETRY_DISABLED = os.getenv("UNPAGE_TELEMETRY_DISABLED", "false").lower() in (
    "1",
    "true",
    "yes",
)


_unpage_telemetry_log_events = os.getenv("UNPAGE_TELEMETRY_LOG_EVENTS", "false").lower() in (
    "1",
    "true",
    "yes",
)


class TunaClient(httpx.AsyncClient):
    BASE_URL = "https://tuna.aptible.com"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(
            *args,
            base_url=self.BASE_URL,
            headers={
                "Content-Type": "application/json; charset=utf-8",
            },
            timeout=httpx.Timeout(
                connect=30,
                read=30,
                write=45,
                pool=15,
            ),
            **kwargs,
        )
        self._user_id = _get_or_create_user_id()
        self._run_id = str(uuid.uuid4())

        # Check if telemetry is disabled globally
        active_config = manager.get_active_profile_config()
        self._telemetry_enabled = active_config.telemetry_enabled and not UNPAGE_TELEMETRY_DISABLED
        if not self._telemetry_enabled:
            # Let the user know that their preference is being respected
            print("Telemetry is disabled", file=sys.stderr)
        if _unpage_telemetry_log_events:
            print("enabled telemetry event logging", file=sys.stderr)

    @property
    def user_id(self) -> str:
        """Unique user_id that is created for each invocation of the unpage program"""
        return self._user_id

    async def send_event(self, event: dict[str, Any]) -> None:
        """Record a telemetry event."""
        if not self._telemetry_enabled:
            return

        try:
            uname = os.uname()
            params = {
                "id": str(uuid.uuid4()),
                "user_id": self.user_id,
                "type": "unpage_telemetry",
                "url": "https://github.com/aptible/unpage",
                "value": json.dumps(
                    {
                        "version": version("unpage"),
                        "github": os.getenv("GITHUB_ACTIONS"),
                        "gitlab": os.getenv("GITLAB_CI"),
                        "travis": os.getenv("TRAVIS"),
                        "circleci": os.getenv("CIRCLECI"),
                        "sysname": uname.sysname,
                        "sysmachine": uname.machine,
                        "sysversion": uname.version,
                        "run_id": self._run_id,
                        **event,
                    }
                ),
            }
            if _unpage_telemetry_log_events:
                rich.print("unpage.telemetry.send_event", {"params": params}, file=sys.stderr)
            response = await self.get(
                "/www/e",
                params=params,
            )
            response.raise_for_status()
        except Exception as e:
            sentry_sdk.capture_exception(e)
            # Don't re-raise the exception (telemetry isn't critical)


client = TunaClient()



import sys

import rich

from unpage.cli.profile._app import profile_app
from unpage.config import manager
from unpage.utils import confirm


@profile_app.command
async def delete(profile_name: str, *, force: bool = False) -> None:
    """Delete a profile.

    Parameters
    ----------
    profile_name
        The name of the profile to delete
    force
        Force deletion even if it's the active profile
    """
    # Check if profile exists first
    available_profiles = manager.list_profiles()
    if profile_name not in available_profiles:
        rich.print(f"[red]Profile '{profile_name}' does not exist[/red]")
        rich.print(f"Available profiles: {', '.join(available_profiles)}")
        sys.exit(1)

    # Prevent deleting the default profile
    if profile_name == "default":
        rich.print("[red]Cannot delete the default profile[/red]")
        sys.exit(1)

    # Check if the profile is the active profile
    active_profile = manager.get_active_profile()
    if profile_name == active_profile and not force:
        rich.print(
            f"[red]Cannot delete active profile '{profile_name}'. Use --force or switch to a different profile first.[/red]"
        )
        sys.exit(1)

    # Confirm deletion unless forced
    if not force:
        confirmed = await confirm(f"Are you sure you want to delete profile '{profile_name}'?")
        if not confirmed:
            rich.print("Deletion cancelled.")
            return

    # Attempt to delete the profile
    manager.delete_profile(profile_name)
    rich.print(f"[green]Profile '{profile_name}' deleted successfully[/green]")

    # If we deleted the active profile, inform user of the switch
    if profile_name == active_profile:
        # Switch to the first profile
        manager.set_active_profile("default")
        rich.print(f"[yellow]Switched to profile '{manager.get_active_profile()}'[/yellow]")



import sys

from rich import print

from unpage.agent.utils import get_agent_template
from unpage.cli.agent._app import agent_app
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import hash_value, prepare_profile_for_telemetry
from unpage.utils import edit_file


@agent_app.command
async def edit(
    agent_name: str,
    /,
    *,
    editor: str | None = None,
) -> None:
    """Edit an existing agent configuration file.

    Parameters
    ----------
    agent_name
        The name of the agent to edit
    editor
        The editor to use to open the agent file; DAYDREAM_EDITOR and EDITOR environment variables also work
    """
    await telemetry.send_event(
        {
            "command": "agent edit",
            "agent_name_sha256": hash_value(agent_name),
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "editor": editor,
        }
    )
    # Get the config directory for the profile
    config_dir = manager.get_active_profile_directory()

    # Build the agent file path
    agent_file = config_dir / "agents" / f"{agent_name}.yaml"

    # If they're editing the default agent and it doesn't exist, create it.
    if agent_name == "default" and not agent_file.exists():
        agent_file = config_dir / "agents" / "default.yaml"
        agent_file.parent.mkdir(parents=True, exist_ok=True)
        agent_file.touch()
        agent_file.write_text(get_agent_template(agent_name))

    # Check if the agent file exists
    if not agent_file.exists():
        print(f"Agent '{agent_name}' not found at {agent_file}")
        print(f"Use 'unpage agent create {agent_name}' to create a new agent.")
        sys.exit(1)

    # Open the file in the user's editor
    try:
        await edit_file(agent_file, editor)
    except ValueError:
        print(
            "[red]No editor specified. Set the $EDITOR environment variable or use --editor option.[/red]"
        )
        print(f"[blue]Please manually open {str(agent_file)!r} in your editor.[/blue]")
        sys.exit(1)



import os
from datetime import UTC, datetime, timedelta
from typing import Any

import questionary
from pydantic import AwareDatetime, BaseModel

from unpage.config import PluginSettings
from unpage.plugins.base import Plugin
from unpage.plugins.mixins import McpServerMixin, tool
from unpage.plugins.papertrail.client import PapertrailClient, PapertrailLogEvent

RESULT_LIMIT = 75 * 1024


class PapertrailSearchResult(BaseModel):
    """Result of a Papertrail log search."""

    truncated: bool = False
    """True when the search results were truncated due to the response size limit or single search time limit"""
    timed_out: bool = False
    """True when the search timed out due to the overall time limit"""
    results: list[PapertrailLogEvent]
    """The log events that were found"""


class PapertrailPlugin(Plugin, McpServerMixin):
    _client: PapertrailClient

    def __init__(
        self,
        *args: Any,
        token: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.token = (
            token
            or os.environ.get("PAPERTRAIL_API_TOKEN", "")
            or os.environ.get("PAPERTRAIL_TOKEN", "")
        )
        if self.token:
            self._client = PapertrailClient(token=self.token)

    async def validate_plugin_config(self) -> None:
        await super().validate_plugin_config()
        await self._client.verify_connection()

    async def interactive_configure(self) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        return {
            "token": await questionary.password(
                "Enter your Papertrail API token",
                default=self.token
                or os.environ.get("PAPERTRAIL_API_TOKEN", "")
                or os.environ.get("PAPERTRAIL_TOKEN", ""),
                instruction="Generate a token with https://docs.unpage.ai/plugins/papertrail#prerequisites",
            ).unsafe_ask_async(),
        }

    @tool()
    async def search_logs(
        self,
        query: str,
        min_time: AwareDatetime,
        max_time: AwareDatetime,
        timeout_seconds: int = 10,
    ) -> PapertrailSearchResult | str:
        """Search Papertrail for logs within a given time range

        Args:
            query (str): The search query.
            min_time (AwareDatetime): The starting time for the range within which to search.
            max_time (AwareDatetime): The ending time for the range within which to search.
            timeout_seconds (int): The maximum seconds to wait for the search to complete. Defaults to 10 seconds.

        Returns:
            PapertrailSearchResult: logs that matched the query and fit within response limit
        """
        if min_time >= max_time:
            return f"min_time must come before max_time {min_time=} {max_time=}"
        logs = []
        truncated = False
        content_length = 0

        class timeoutTracker:
            timed_out: bool = False
            start_time: AwareDatetime = datetime.now(UTC)
            time_out: timedelta = timedelta(seconds=timeout_seconds)

            def under_time_out(self, _: AwareDatetime | None) -> bool:
                self.timed_out = (datetime.now(UTC) - self.start_time) > self.time_out
                return not self.timed_out

        tt = timeoutTracker()
        async for log in self._client.search(
            query=query,
            min_time=min_time,
            max_time=max_time,
            continue_search=tt.under_time_out,
        ):
            content_length += len(log.model_dump_json())
            if content_length >= RESULT_LIMIT:
                truncated = True
                break
            logs.append(log)
        return PapertrailSearchResult(
            results=logs,
            truncated=truncated,
            timed_out=tt.timed_out,
        )



import sys
from pathlib import Path

from unpage.agent.utils import get_agent_template
from unpage.config import manager
from unpage.utils import confirm


async def create_agent(agent_name: str, overwrite: bool, template: str) -> Path:
    # Create the default YAML content
    try:
        agent_template = get_agent_template(template)
    except FileNotFoundError:
        print(
            f"Template '{template}' not found at {Path(__file__).parent / 'templates' / f'{template}.yaml'}",
            file=sys.stderr,
        )
        sys.exit(1)

    # Get the config directory for the active profile
    config_dir = manager.get_active_profile_directory()

    # Create the agents directory if it doesn't exist
    agents_dir = config_dir / "agents"
    agents_dir.mkdir(exist_ok=True)

    # Create the agent file path
    agent_file = agents_dir / f"{agent_name}.yaml"

    # Check if the agent file already exists
    if agent_file.exists():
        if overwrite:
            print(f"Overwriting agent '{agent_name}' at {agent_file}")
        else:
            print(f"Agent '{agent_name}' already exists at {agent_file}")
            if not await confirm("Do you want to overwrite it?"):
                sys.exit(1)

    # Write the YAML content to the file
    agent_file.write_text(agent_template, encoding="utf-8")

    print(f"Created agent configuration at {agent_file}")
    return agent_file



import os
import signal
import sys

from unpage.cli.graph._app import graph_app
from unpage.cli.graph._background import cleanup_pid_file, get_pid_file, is_process_running
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@graph_app.command
async def stop() -> None:
    """Stop running graph build"""
    active_profile = manager.get_active_profile()
    await telemetry.send_event(
        {
            "command": "graph stop",
            **prepare_profile_for_telemetry(active_profile),
        }
    )

    pid_file = get_pid_file()

    if not pid_file.exists():
        print("No graph build running")
        sys.exit(1)

    try:
        pid = int(pid_file.read_text().strip())
        if not is_process_running(pid):
            cleanup_pid_file()
            print("No graph build running")
            sys.exit(1)

        print(f"Stopping graph build (PID: {pid})...")
        os.kill(pid, signal.SIGTERM)
        print("Graph build stopped successfully")
        cleanup_pid_file()
    except (ValueError, ProcessLookupError):
        cleanup_pid_file()
        print("No graph build running")
        sys.exit(1)



import asyncio
import os
import shlex
import sys
import time
from collections import Counter

import anyio

from unpage.cli.graph._app import graph_app
from unpage.cli.graph._background import (
    check_and_create_lock,
    cleanup_pid_file,
    create_pid_file,
    get_log_file,
)
from unpage.config import manager
from unpage.knowledge import Graph
from unpage.plugins import PluginManager
from unpage.plugins.mixins import KnowledgeGraphMixin
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@graph_app.command
async def build(
    *,
    interval: int | None = None,
    background: bool = False,
) -> None:
    """Build a knowledge graph for your cloud infrastructure

    Parameters
    ----------
    interval
        Rebuild the graph continuously, pausing for the specified seconds between builds
    background
        Run in background and return immediately
    """
    # Check if already running
    if not check_and_create_lock():
        print("Graph build already running")
        print("Use 'unpage graph stop' to stop it if needed")
        sys.exit(1)

    if background:
        # Build command args for subprocess
        cmd = [a for a in sys.argv if a != "--background"]

        # Set up logging
        log_file = get_log_file()
        log_file.parent.mkdir(parents=True, exist_ok=True)

        # Start subprocess with logging
        with log_file.open("w") as f:
            f.write("Starting graph build in background\n")
            f.write(f"Command: {' '.join(cmd)}\n")
            f.write("=" * 50 + "\n\n")
            f.flush()

            await asyncio.create_subprocess_shell(
                shlex.join(cmd), stdout=f, stderr=asyncio.subprocess.STDOUT, start_new_session=True
            )

        print("Graph building started in background")
        print("Check progress: unpage graph logs --follow")
        print("Stop with: unpage graph stop")
        return

    async def _build_graph() -> None:
        await telemetry.send_event(
            {
                "command": "graph build - starting",
                **prepare_profile_for_telemetry(manager.get_active_profile()),
                "interval": interval,
                "background": background,
            }
        )
        print("Building graph...")

        start_time = time.perf_counter()

        graph = Graph()
        config = manager.get_active_profile_config()
        output_path = (manager.get_active_profile_directory() / "graph.json").resolve()
        plugin_manager = PluginManager(config)

        async with anyio.create_task_group() as tg:
            for plugin in plugin_manager.get_plugins_with_capability(KnowledgeGraphMixin):
                print(f"Populating graph with the {plugin.name} plugin...")
                tg.start_soon(plugin.populate_graph, graph)

        await graph.infer_edges()

        print(f"Saving graph to {output_path!s}...")
        await graph.save(output_path)

        end_time = time.perf_counter()
        total_time = end_time - start_time
        edge_counts = Counter(
            [edge.properties["relationship_type"] async for edge in graph.iter_edges()]
        )
        node_counts = Counter([node.node_type async for node in graph.iter_nodes()])

        print("=== Summary ===")

        print("Edges:")
        for relationship_type, count in edge_counts.items():
            print(f"  {relationship_type}: {count}")
        print("Nodes:")
        for node_type, count in node_counts.items():
            print(f"  {node_type}: {count}")

        print(f"Graph built in {total_time:.2f} seconds")
        print(f"Graph saved to {output_path!s}")
        print("=== End Summary ===")

        await telemetry.send_event(
            {
                "command": "graph build - finished",
                **prepare_profile_for_telemetry(manager.get_active_profile()),
                "duration_seconds": total_time,
                "node_counts": node_counts,
                "edge_counts": edge_counts,
            }
        )

    try:
        create_pid_file(os.getpid())

        if interval:
            while True:
                await _build_graph()
                print(f"Sleeping for {interval} seconds before next build...")
                await anyio.sleep(interval)
        else:
            await _build_graph()
    finally:
        cleanup_pid_file()



from unpage.plugins.aptible.nodes.base import AptibleNode


class AptibleVhost(AptibleNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["virtual_domain"],
            self.raw_data["external_host"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            self.raw_data["security_group_id"],
            (self.raw_data["elastic_load_balancer_name"], "receives_traffic_from"),
            (self.raw_data["application_load_balancer_arn"], "receives_traffic_from"),
            self.raw_data["acme_dns_challenge_host"],
        ]



from datetime import datetime

from pydantic import AwareDatetime, BaseModel


class LogLine(BaseModel):
    time: datetime
    log: str


class Observation(BaseModel):
    node_id: str
    observation_type: str
    data: dict[AwareDatetime, float]



import sys

from rich import print

from unpage.agent.analysis import AnalysisAgent
from unpage.cli.agent._app import agent_app
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@agent_app.command
async def route(
    payload: str | None = None,
    /,
    *,
    debug: bool = False,
) -> None:
    """Determine which agent will be used to analyze the given payload.

    A payload can be passed as an argument or piped to stdin.

    Parameters
    ----------
    payload
        The alert payload to analyze. Alternatively, you can pipe the payload to stdin.
    debug
        Enable debug mode to print the history of the routing agent.
    """
    await telemetry.send_event(
        {
            "command": "agent route",
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "debug": debug,
        }
    )
    # Read data from stdin if it's being piped to us.
    if not sys.stdin.isatty():
        if payload is not None:
            print("[red]Cannot pass a payload argument when piping data to stdin.[/red]")
            sys.exit(1)
        data = sys.stdin.read().strip()
    else:
        # Otherwise, use the payload argument.
        data = payload

    if not data:
        print("[red]No payload provided.[/red]")
        print(
            "[bold]Pass an alert payload as an argument or pipe the payload data to stdin.[/bold]"
        )
        sys.exit(1)

    # Run the analysis with the specific agent
    try:
        analysis_agent = AnalysisAgent()
        result = await analysis_agent.acall(payload=data, route_only=True)
        if debug:
            print("\n\n===== DEBUG OUTPUT =====\n")
            analysis_agent.inspect_history(n=1000)
            print("\n===== END DEBUG OUTPUT =====\n\n")
        print(result)
    except Exception as ex:
        print(f"[red]Analysis failed:[/red] {ex}")
        sys.exit(1)



from rich import print

from unpage.agent.utils import get_agent_templates
from unpage.cli.agent._app import agent_app
from unpage.telemetry import client as telemetry


@agent_app.command
async def templates() -> None:
    """List the available agent templates."""
    await telemetry.send_event(
        {
            "command": "agent templates",
        }
    )
    print("Available agent templates:")
    for template in sorted(get_agent_templates()):
        print(f"* {template}")



import os
import socket
import statistics
import time

import dns.resolver
import requests
from dns.exception import Timeout as DnsTimeout
from scapy.all import sr1
from scapy.layers.inet import ICMP, IP, UDP

from unpage.plugins import Plugin
from unpage.plugins.mixins import McpServerMixin, tool


class NetworkingPlugin(Plugin, McpServerMixin):
    """A plugin for providing common networking tools."""

    @tool()
    def ping(self, host: str) -> str:
        """Ping a host and return the result."""
        for segment in host.split("."):
            if len(segment) > 63:
                raise ValueError(
                    f"'{host}' host has a segment longer than 63 characters, which is not supported by dns"
                )
        try:
            target_ip = socket.gethostbyname(host)
        except socket.gaierror:
            return f"ping: cannot resolve {host}: Unknown host"
        result = f"PING {host} ({target_ip}): 56 data bytes\n"
        packets_transmitted = 0
        packets_received = 0
        rtt_values = []
        for i in range(4):
            packet = IP(dst=target_ip) / ICMP(id=os.getpid() & 0xFFFF, seq=i)
            reply = sr1(packet, verbose=0, timeout=1)
            packets_transmitted += 1
            if reply and reply.haslayer(ICMP) and reply[ICMP].type == 0:
                packets_received += 1
                rtt = (reply.time - packet.sent_time) * 1000  # pyright: ignore[reportOperatorIssue]
                rtt_values.append(rtt)
                result += f"64 bytes from {reply[IP].src} icmp_seq={i} ttl={reply[IP].ttl} time={rtt:.3f} ms\n"
            else:
                result += f"Request timeout for icmp_seq {i}\n"
            time.sleep(1)
        result += f"\n--- {host} ping statistics ---\n"
        result += (
            f"{packets_transmitted} packets transmitted, {packets_received} packets received, "
            f"{((packets_transmitted - packets_received) / packets_transmitted * 100):.1f}% packet loss\n"
        )
        if rtt_values:
            min_rtt = min(rtt_values)
            avg_rtt = statistics.mean(rtt_values)
            max_rtt = max(rtt_values)
            stddev_rtt = statistics.stdev(rtt_values) if len(rtt_values) > 1 else 0.0
            result += f"round-trip min/avg/max/stddev = {min_rtt:.3f}/{avg_rtt:.3f}/{max_rtt:.3f}/{stddev_rtt:.3f} ms\n"
        return result

    @tool()
    def dig(self, hostname: str, record_type: str = "A") -> str:
        """Use the 'dig' command to resolve a hostname and return the result."""
        result = ""
        try:
            resolver = dns.resolver.Resolver()
            answers = resolver.resolve(hostname, record_type)
            for rdata in answers:
                result += str(rdata) + "\n"
        except dns.resolver.NXDOMAIN:
            result = f"Host '{hostname}' not found."
        except dns.resolver.NoAnswer:
            result = f"No record found for '{hostname}' with type '{record_type}'."
        except DnsTimeout:
            result = "Timeout while querying DNS server."

        return result

    @tool()
    def traceroute(self, destination: str, max_hops: int = 30) -> str:
        """Trace the route to a destination and return the result."""
        result = ""
        for ttl in range(1, max_hops + 1):
            packet = IP(dst=destination, ttl=ttl) / UDP(dport=33434)
            reply = sr1(packet, verbose=0, timeout=1)
            if reply is None:
                result += f"{ttl}\t* * * Request timed out.\n"
            elif reply.haslayer(ICMP) and reply[ICMP].type == 11:
                result += f"{ttl}\t{reply.src}\n"
            elif reply.haslayer(ICMP) and reply[ICMP].type == 3:
                result += f"{ttl}\t{reply.src} Destination reached.\n"
                break
            else:
                result += f"{ttl}\t{reply.src} Other reply type.\n"

        return result

    @tool()
    def request_url(self, url: str) -> str:
        """Request a URL with the GET method and return the result."""
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        return response.text



import rich

from unpage.cli.mcp.client.logs.app import client_logs_app
from unpage.telemetry import client as telemetry


@client_logs_app.command()
async def claude() -> None:
    """Show logs for Claude Desktop Unpage MCP Server"""
    rich.print("Logs are available at:")
    rich.print("~/Library/Logs/Claude/mcp.log")
    rich.print("~/Library/Logs/Claude/mcp-server-unpage.log")
    await telemetry.send_event(
        {
            "command": "mcp client logs claude",
        }
    )



from typing import Any

from pydantic import AwareDatetime, BaseModel


class DatadogLogEventAttributes(BaseModel):
    timestamp: AwareDatetime
    host: str
    service: str
    status: str
    tags: list[str]
    attributes: dict[str, Any]


class DatadogLogEvent(BaseModel):
    """Datadog log event"""

    """ID of the Log"""
    id: str
    """Type of event (always log)"""
    type: str = "log"
    """All other log attributes"""
    attributes: DatadogLogEventAttributes


class DatadogLogSearchResult(BaseModel):
    """Result of a DataDog log search."""

    """True when the search results were truncated due to the response size limit or single search time limit"""
    truncated: bool = False
    """The log events that were found"""
    results: list[DatadogLogEvent]



from unpage.knowledge.graph import Node


class KubernetesBaseNode(Node):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data.get("metadata", {}).get("uid"),
            self.raw_data.get("metadata", {}).get("name"),
        ]

    async def get_reference_identifiers(self) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            *[
                (owner_ref.get("uid", None), "owned_by")
                for owner_ref in self.raw_data.get("metadata", {}).get("ownerReferences", [])
            ],
        ]



from typing import Any

from fastmcp import FastMCP
from fastmcp.contrib.mcp_mixin import (
    MCPMixin,
    mcp_prompt,
    mcp_resource,
    mcp_tool,
)

from unpage.plugins import PluginCapability


class McpServerMixin(MCPMixin, PluginCapability):
    """Capability for registering tools, resources, and prompts with the MCP server."""

    def get_mcp_server(self) -> FastMCP[Any]:
        mcp = FastMCP[Any](self.name)
        self.register_all(mcp)
        return mcp


prompt = mcp_prompt
resource = mcp_resource
tool = mcp_tool



import re
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Any, TypeVar

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
)

from unpage.utils import camel_to_snake, classproperty

if TYPE_CHECKING:
    from unpage.knowledge import Edge, Graph

NODE_REGISTRY: dict[str, type["Node"]] = {}

_NodeType = TypeVar("_NodeType", bound="Node")


class Node(BaseModel):
    node_id: str
    raw_data: dict[str, Any] = Field(default_factory=dict)

    _graph: "Graph"

    model_config = ConfigDict(ignored_types=(classproperty,))

    def __init__(self, *args: Any, _graph: "Graph", **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._graph = _graph

    @classproperty
    def _node_source(cls) -> str:
        match = re.match(r"^unpage\.plugins\.([^\.]+)\..*$", cls.__module__)
        if not match:
            raise ValueError(f"Node source not found for {cls.__module__}")
        return match.group(1)

    @classproperty
    def _node_type(cls) -> str:
        return camel_to_snake(cls.__name__)

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        """Register the node type in the registry."""
        super().__pydantic_init_subclass__(**kwargs)

        key = f"{cls._node_source}:{cls._node_type}"

        if key in NODE_REGISTRY:
            raise ValueError(f"Node type {key} already registered")

        NODE_REGISTRY[key] = cls

    @computed_field
    @property
    def node_source(self) -> str:
        return self._node_source

    @computed_field
    @property
    def node_type(self) -> str:
        return self._node_type

    @computed_field
    @property
    def nid(self) -> str:
        """Return the canonical identifier for the node."""
        return ":".join(filter(bool, (self.node_source, self.node_type, self.node_id)))

    async def get_identifiers(self) -> list[str | None]:
        """Return a list of alternative identifiers for the node."""
        return []

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        """Return a list of values from this node that may be references to other nodes.

        The values may be the referenced value, or tuples of the referenced value and the relationship type.
        """
        return []

    async def iter_successors(self) -> AsyncIterator["Node"]:
        """Return a list of nodes that are successors of this node."""
        async for successor in self._graph.iter_successors(self):
            yield successor

    async def iter_predecessors(self) -> AsyncIterator["Node"]:
        """Return a list of nodes that are predecessors of this node."""
        async for predecessor in self._graph.iter_predecessors(self):
            yield predecessor

    async def iter_neighbors(self) -> AsyncIterator["Node"]:
        """Return a list of nodes that are neighbors of this node."""
        async for neighbor in self._graph.iter_neighbors(self):
            yield neighbor

    async def iter_neighboring(
        self,
        only_type: type[_NodeType] | tuple[type[_NodeType], ...],
    ) -> AsyncIterator[_NodeType]:
        """Return a list of nodes that are neighbors of this node and are of a specific type."""
        async for neighbor in self._graph.iter_neighboring(self, only_type):
            yield neighbor

    async def iter_neighborhood_edges(self, max_depth: int = 1) -> AsyncIterator["Edge"]:
        """Return a list of edges that are neighbors of this node."""
        async for edge in self._graph.iter_neighborhood_edges(self, max_depth):
            yield edge



import sys
import warnings

from fastmcp import settings as fastmcp_settings

from unpage import mcp
from unpage.cli.mcp._app import mcp_app
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@mcp_app.command
async def start(
    *,
    disable_sse: bool = False,
    disable_stdio: bool = False,
    disable_http: bool = False,
    http_host: str = fastmcp_settings.host,
    http_port: int = fastmcp_settings.port,
) -> None:
    """Start the Unpage MCP Server

    Parameters
    ----------
    disable_sse
        Disable the HTTP transport for the MCP Server (deprecated, use --disable-http instead)
    disable_stdio
        Disable the stdio transport for the MCP Server
    disable_http
        Disable the HTTP transport for the MCP Server
    http_host
        The host to bind the HTTP transport to
    http_port
        The port to bind the HTTP transport to
    """
    # Deprecate --disable-sse in favor of --disable-http
    if "--disable-sse" in sys.argv:
        disable_http = disable_sse
        warnings.warn(
            "The `--disable-sse` argument is deprecated. Use `--disable-http` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    await telemetry.send_event(
        {
            "command": "mcp start",
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "disable_sse": disable_sse,
            "disable_stdio": disable_stdio,
            "disable_http": disable_http,
            "http_host": http_host,
            "http_port": http_port,
        }
    )
    await mcp.start(
        disable_stdio=disable_stdio,
        disable_http=disable_http,
        http_host=http_host,
        http_port=http_port,
    )



from cyclopts import App

from unpage.cli.mcp.client._app import client_app
from unpage.cli.mcp.tools._app import tools_app

mcp_app = App(help="MCP tool commands")
mcp_app.command(tools_app, name="tools")
mcp_app.command(client_app, name="client")



import base64
import json
import os
import time
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from datetime import UTC, datetime
from pathlib import Path
from typing import Any

import anyio
import httpx
from pydantic import AwareDatetime

from unpage.models import Observation
from unpage.utils import as_completed


class AptibleClient(httpx.AsyncClient):
    """A client for the Aptible API."""

    BASE_API_URL = "https://api.aptible.com"
    BASE_AUTH_URL = "https://auth.aptible.com"

    TOKENS_PATH = Path("~/.aptible/tokens.json").expanduser()

    def __init__(
        self,
        api_key: str | None = None,
        connection_limit: int = 100,
        base_url: str = BASE_API_URL,
    ) -> None:
        self._jwt = api_key

        super().__init__(
            base_url=base_url,
            headers={
                "Content-Type": "application/json; charset=utf-8",
            },
            timeout=httpx.Timeout(None),
            limits=httpx.Limits(max_connections=connection_limit),
        )

    async def request(self, *args: Any, **kwargs: Any) -> httpx.Response:
        return await super().request(
            *args,
            **{
                **kwargs,
                "headers": {
                    # Inject the API key into the headers.
                    "Authorization": f"Bearer {await self.get_api_key()}",
                    **(kwargs["headers"] or {}),
                },
            },
        )

    async def get_api_key(self) -> str:
        # TODO: Check for expiry and refresh if necessary
        if not self._jwt or await self.jwt_expired(self._jwt):
            if "APTIBLE_EMAIL" in os.environ and "APTIBLE_PASSWORD" in os.environ:
                self._jwt = await self.authorize(
                    email=os.environ["APTIBLE_EMAIL"],
                    password=os.environ["APTIBLE_PASSWORD"],
                )
            elif self.TOKENS_PATH.exists():
                tokens = json.loads(self.TOKENS_PATH.read_text().strip())
                self._jwt = tokens[self.BASE_AUTH_URL]
            else:
                raise ValueError("No API key found")
        return self._jwt

    async def authorize(self, email: str, password: str) -> str:
        async with httpx.AsyncClient(base_url=self.BASE_AUTH_URL) as auth_client:
            resp = await auth_client.post(
                f"{self.BASE_AUTH_URL}/tokens",
                json={
                    "username": email,
                    "password": password,
                    "grant_type": "password",
                    "scope": "manage",
                    "expires_in": 3600,
                },
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
            )
            # TODO: Special error messaging for 401s caused by MFA requirements.
            resp.raise_for_status()
            return resp.json()["access_token"]

    async def jwt_expired(self, jwt: str) -> bool:
        # Decode the JWT token
        decoded = jwt.split(".")[1]
        # Add padding if needed for base64 decoding
        decoded += "=" * ((4 - len(decoded) % 4) % 4)
        decoded = base64.b64decode(decoded).decode()
        decoded = json.loads(decoded)
        # Check if the JWT is expired, or will expire in the next 60 seconds.
        return int(decoded["exp"]) < (time.time() - 60)

    async def validate_auth(self) -> None:
        resp = await self.get(f"{self.BASE_AUTH_URL}/organizations")
        try:
            resp.raise_for_status()
        except Exception as ex:
            raise ValueError(
                "Authentication invalid for Aptible API. Try running `aptible login` in another terminal, then try again here"
            ) from ex

    async def paginate(self, url: str) -> AsyncGenerator[dict[str, Any], None]:
        while url:
            response = await self.get(url)
            response.raise_for_status()
            for collection in response.json().get("_embedded", {}).values():
                for item in collection:
                    yield item
            url = response.json().get("_links", {}).get("next", {}).get("href")

    @asynccontextmanager
    async def concurrent_paginator(
        self, url: str
    ) -> AsyncGenerator[tuple[AsyncGenerator[dict[str, Any], None], int], None]:
        """Retrieve all results in parallel.

        Fetches the first page of results to get pagination information, and
        then fetches all subsequent pages in parallel, yielding items as they
        are available. Does NOT maintain result order.
        """
        first_page_response = await self.get(url)
        first_page_response.raise_for_status()

        first_page = first_page_response.json()
        total_count = first_page["total_count"]
        per_page = first_page["per_page"]
        total_pages = total_count // per_page

        async def item_generator() -> AsyncGenerator[dict[str, Any], None]:
            # Yield all results from the first page
            for collection in first_page["_embedded"].values():
                collection = collection if isinstance(collection, list) else [collection]
                for item in collection:
                    yield item

            # Fetch all subsequent pages in parallel
            pages = (self.get(f"{url}?page={p}") for p in range(1, total_pages + 1))
            try:
                async with anyio.create_task_group() as tg:
                    for page in as_completed(tg, pages):
                        try:
                            response = await page
                            response.raise_for_status()
                            for collection in response.json().get("_embedded", {}).values():
                                collection = (
                                    collection if isinstance(collection, list) else [collection]
                                )
                                for item in collection:
                                    yield item
                        except Exception as e:
                            print(f"Error processing page from {url}: {e}")
            except Exception as e:
                print(f"Error in concurrent iterator for {url}: {e}")

        yield (item_generator(), total_count)


class AptibleMetricsClient(AptibleClient):
    """A client for the Aptible Metrics API."""

    METRICS_BASE_URL = "https://metrictunnel-nextgen.aptible.com"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs, base_url=self.METRICS_BASE_URL)

    async def get_metrics(
        self,
        node_id: str,
        container_ids: list[int | str],
        metric: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation]:
        ts = int(time_range_end.timestamp() * 1000)

        response = await self.get(
            f"/proxy/{':'.join(str(cid) for cid in container_ids)}",
            params={
                "horizon": "1h",
                "ts": ts,
                "metric": metric,
                "requestedTicks": 300,
            },
        )
        data = response.json()

        if "error" in data:
            print(f"Error retrieving metrics: {data['error']}")
            return []

        dates: list[AwareDatetime] = []
        value_columns: dict[str, list[float]] = {}

        # Split columns into dates and values
        for first_col, *other_cols in data.get("columns", []):
            if first_col == "time_0":
                dates = [
                    datetime.strptime(d, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=UTC)
                    for d in other_cols
                ]
            else:
                value_columns[first_col] = other_cols

        # For each series, zip its values with the dates
        result = [
            Observation(
                node_id=node_id,
                observation_type=metric,
                data={
                    date: value
                    for date, value in zip(dates, values, strict=True)
                    if value is not None
                },
            )
            for values in value_columns.values()
        ]

        return result



from typing import Any

from pydantic import AwareDatetime, BaseModel


class PagerDutyService(BaseModel):
    id: str
    type: str
    summary: str
    html_url: str


class PagerDutyAssignee(BaseModel):
    id: str
    type: str
    summary: str
    html_url: str


class PagerDutyAssignment(BaseModel):
    at: AwareDatetime
    assignee: PagerDutyAssignee


class PagerDutyAlertCounts(BaseModel):
    all: int
    triggered: int
    resolved: int


class PagerDutyIncidentType(BaseModel):
    name: str


class PagerDutyEscalationPolicy(BaseModel):
    id: str
    type: str
    summary: str
    html_url: str


class PagerDutyTeam(BaseModel):
    id: str
    html_url: str


class PagerDutyIncident(BaseModel):
    incident_number: int
    title: str
    description: str
    created_at: AwareDatetime
    updated_at: AwareDatetime
    status: str
    incident_key: str | None = None
    service: PagerDutyService
    assignments: list[PagerDutyAssignment] | None = None
    assigned_via: str | None = None
    last_status_changed_at: AwareDatetime | None = None
    resolved_at: AwareDatetime | None = None
    alert_counts: PagerDutyAlertCounts
    is_mergeable: bool
    incident_type: PagerDutyIncidentType
    escalation_policy: PagerDutyEscalationPolicy
    teams: list[PagerDutyTeam] | None = None
    impacted_services: list[PagerDutyService] | None = None
    pending_actions: list | None = None
    acknowledgements: list | None = None
    basic_alert_grouping: Any | None = None
    alert_grouping: Any | None = None
    last_status_change_by: Any
    incident_responders: list | None = None
    responder_requests: list | None = None
    subscriber_requests: list | None = None
    urgency: str
    id: str
    type: str
    summary: str
    html_url: str


class PagerDutyIncidentPayload(BaseModel):
    incident: PagerDutyIncident



from collections.abc import AsyncGenerator, Callable
from typing import Any
from urllib.parse import parse_qs, urlparse

import httpx
from pydantic import AwareDatetime, BaseModel


class SolarWindsLogEvent(BaseModel):
    time: AwareDatetime
    """time that SolarWinds received the log event (ISO 8601 timestamp)"""

    message: str
    """log event message"""


class ResultPageInfo(BaseModel):
    """Pagination links"""

    prevPage: str | None = None
    nextPage: str | None = None


class SearchResult(BaseModel):
    """Search result from SolarWinds API"""

    logs: list[SolarWindsLogEvent]
    """An array of hashes of log events (one hash per event)"""

    pageInfo: ResultPageInfo


class SolarWindsClient(httpx.AsyncClient):
    """SolarWinds API client

    https://api.na-01.cloud.solarwinds.com/v1/#/logs/searchLogs
    """

    BASE_URL_FORMAT = "https://api.{datacenter}.cloud.solarwinds.com"

    def __init__(self, token: str, datacenter: str, connection_limit: int = 10) -> None:
        self._token = token
        base_url = self.BASE_URL_FORMAT.format(datacenter=datacenter)
        super().__init__(
            base_url=base_url,
            headers={
                "Authorization": f"Bearer {self._token}",
            },
            timeout=httpx.Timeout(60),
            limits=httpx.Limits(max_connections=connection_limit),
        )

    async def verify_connection(self) -> None:
        response = await self.get("/v1/logs", params={"pageSize": 1})
        response.raise_for_status()
        SearchResult.model_validate(response.json())

    @staticmethod
    def extract_skip_token(path: str) -> str | None:
        """Extract the skipToken parameter from the search result.

        Args:
          path: The URL path string, returned by the Solarwinds search result
          e.g. "/v1/logs?filter=test&skipToken=ABC123&direction=backward"

        Returns:
          The skipToken value, or None
        """
        parsed_url = urlparse(path)
        query_params = parse_qs(parsed_url.query)
        skip_tokens = query_params.get("skipToken", [])
        return skip_tokens[0] if skip_tokens else None

    async def search(
        self,
        query: str,
        min_time: AwareDatetime,
        max_time: AwareDatetime,
        page_size: int = 1000,
        continue_search: Callable[[], bool] | None = None,
    ) -> AsyncGenerator[SolarWindsLogEvent, None]:
        """Search logs with pagination.

        Args:
            query: The search filter string
            min_time: Start time for the search
            max_time: End time for the search
            page_size: Number of results per page

        Yields:
            Log events matching the search criteria
        """
        skip_token = None

        while True:
            params: dict[str, Any] = {
                "filter": query,
                "startTime": min_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "endTime": max_time.strftime("%Y-%m-%dT%H:%M:%SZ"),
                "pageSize": page_size,
                "direction": "backward",
            }

            if skip_token:
                params["skipToken"] = skip_token

            response = await self.get("/v1/logs", params=params)
            response.raise_for_status()
            result = SearchResult(**response.json())

            # Yield all logs in the current page
            for log in result.logs:
                yield log

            next_page = getattr(result.pageInfo, "nextPage", None)
            new_token = self.extract_skip_token(next_page) if next_page else None

            # If there are no more new results, we can break early.
            if new_token in (None, skip_token):
                break

            skip_token = new_token
            if continue_search is None or not continue_search():
                break



import asyncio
import os
import shlex
import sys
from pathlib import Path
from typing import Any

import questionary
import rich

from unpage.cli._app import app
from unpage.config import Config, PluginConfig, manager
from unpage.plugins.base import PluginManager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry
from unpage.utils import Choice, checkbox, confirm


def _resolve_default_use_uv_run() -> bool:
    try:
        cwd = Path.cwd()
        git_dir = None
        for parent in [cwd, *cwd.parents[:10]]:
            if (parent / ".git").exists():
                git_dir = parent
                break
        if git_dir is None:
            return False
        repo_name = git_dir.name
        return repo_name == "unpage"
    except Exception:
        return False


_default_use_uv_run = _resolve_default_use_uv_run()


async def _send_event(step: str, extra_params: dict[Any, Any] | None = None) -> None:
    await telemetry.send_event(
        {
            "command": "configure",
            "step": step,
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            **(extra_params if extra_params else {}),
        }
    )


@app.command
async def configure(
    *,
    use_uv_run: bool = _default_use_uv_run,
) -> None:
    """Setup unpage including all plugins!

    Parameters
    ----------
    use_uv_run
        Use uv run instead of uvx to start the Unpage MCP server (useful for developing Unpage)
    """
    await _send_event(
        "start",
        extra_params={
            "use_uv_run": use_uv_run,
        },
    )
    welcome_to_unpage()
    await _configure_intro()
    cfg = _initial_config()
    await _select_plugins_to_enable_disable(cfg)
    cfg.save()
    await _send_event("config_saved")
    rich.print("")
    await _configure_plugins(cfg)
    await _send_event("plugins_configured")
    cfg.save()
    await _send_event("config_saved_2")
    rich.print("")
    await _suggest_building_graph(use_uv_run)


def welcome_to_unpage() -> None:
    rich.print("""
                      [deep_sky_blue1]                        -@@@@@%%%%+       [/deep_sky_blue1]
                      [deep_sky_blue1]                             :+@@@*       [/deep_sky_blue1]
                      [deep_sky_blue1]             %@@@@@@%       +@@#-         [/deep_sky_blue1]
                      [deep_sky_blue1]                 %@*     :%@@#            [/deep_sky_blue1]
                      [deep_sky_blue1]     =*%%-     +@%:     #%@@#+.           [/deep_sky_blue1]
                      [deep_sky_blue1]   -=-=%#     #@@+---    -=+*#%%@%        [/deep_sky_blue1]
                      [deep_sky_blue1]     .%+      +##****                     [/deep_sky_blue1]
                      [deep_sky_blue1]     %@@@@=                               [/deep_sky_blue1]
                                                .-+#%%%%%%#:
                              :--==+#%%@%#*=-::..      ..  :%+
                .=+**##%%%%*=:.      .:-+***##%%%*+++++:%@%: =#
            *%#+:       :+#%%%%%%%##*+==*#%%###*+:     +#+@@# =%
         .##: :*%%%%#*###+++*#*++==-.                   .*#@@* **
       -%@# -%@@%*##=:.                                  =*%@@+ %=
      %@@@+ #@@@=%        [bright_green]                        .%  [/bright_green]    *+@@@: %.
     %@@@@# *@@@%#:       [bright_green]                   +*   =@. [/bright_green]    :%*@@@.:*
    -@@@@@%..@@@@=+       [bright_green]  +:    %=          =%##%=  [/bright_green]     +##@@%.++
    -@@@@@@# *@@@%+:      [bright_green]  :%*. *%.                  [/bright_green]      %#%@@*.#:
     %@@@@@@- %@@%*+      [bright_green]    .+*-           :+       [/bright_green]       **@@%.-@
     =@@@@@@% :%@@*+.     [bright_green]            -#-..:+%-       [/bright_green]       %#%@@* *#
      #@@@@@@* +@@%+*     [bright_green]               --:          [/bright_green]   .=*%#%@@@@= #=
      =@@@@@@%..%@@#*+                        .-+#%%%%###%%@@@@@@@@@: %.
       #@@@@@@+ *@@@-#:           .:-+*###**#**+*#%@%%#+=-+%.    .+%@ -#
       -%@@@@@%: @@@@-+*+++**##*+++*#**#%@@@@@@@%+..:-=*#- #-.+.=#:-@* *=
        #@@@%@@# =@@@@@%%%%%%%@@@@@@%%%%@@@@@@@+.%%%.   @% =# .  .+ #%-:%
         @@@@%%@= %@@@%***+=--==+*%%%%%#+:-%@@@* +%%+-. %@: % -=++ .@%- %.
         -@@* .%@ .%@ +%%+-:.         +@@@* %@%%#=.   ..    #=   =#@@# :%.
          #*   #@# -@=.%+        -+*##%@%+ .@@@@@%%@@@@@@@@@@@@@@%%#.  **
          %   -@@@- *@* .*#######*=     .*%%@@@@@@@@%%%##*-       .=*+%*
         =+   %@@@#  %@@@*:..::-=*%@@%@@@@%#++=-:    .:-=+#%@@@@%@@@@%=
         #-   @@@%@#  -%@@@@@@@%#+-..     :=+**#%%@@@@@@@@@@@@@@@@@@#
         #-   @@@@%@@%-        :=#%%@@@@@@@@@@@*      -%@@@@@@*   -%
         +#   #@@@@@@%%@@@@@@@@@@%@%*      *%@@@@@#**#=  .+%%- =%*-%
          #+   :+%@@%@@@@@@@@@@@@@@@@%@%*:   =%@#     #%#: #-.%@@%*%
          :%        #@@@@@@@@@@@@@@%%##%@@%=  -%. *%+ #@@@@# %@@%*%:
           +%   +  + *@%###+-.           +%@%%%= *@@=-# -%@@@@@%*%-
             +%- -*.#%+                    =%@@=*@@#.*    #@@@@%*
               :**=-.                        *@@@@#+*        .
                                               *%%+

              ██╗   ██╗███╗   ██╗██████╗  █████╗  ██████╗ ███████╗
              ██║   ██║████╗  ██║██╔══██╗██╔══██╗██╔════╝ ██╔════╝
              ██║   ██║██╔██╗ ██║██████╔╝███████║██║  ███╗█████╗
              ██║   ██║██║╚██╗██║██╔═══╝ ██╔══██║██║   ██║██╔══╝
              ╚██████╔╝██║ ╚████║██║     ██║  ██║╚██████╔╝███████╗
               ╚═════╝ ╚═╝  ╚═══╝╚═╝     ╚═╝  ╚═╝ ╚═════╝ ╚══════╝

> Welcome to Unpage!""")


async def _configure_intro() -> None:
    rich.print(""">
> This interactive tool will setup Unpage plugins so you'll be ready to:
>
>   • Build a knowlege graph of your infrastructure
>   • Use Unpage's MCP server tools to interact with your graph, logs, and metrics
>   • Write your own agents using everything above!
> """)
    rich.print("> Ready to start?")
    rich.print("")
    await questionary.press_any_key_to_continue().unsafe_ask_async()


def _initial_config() -> Config:
    default_config = manager.get_empty_config(manager.get_active_profile())
    try:
        existing_config = manager.get_active_profile_config()
    except Exception:
        existing_config = default_config
    plugin_settings: dict[str, PluginConfig] = {}
    for plugin_name in default_config.plugins:
        plugin_settings[plugin_name] = PluginConfig(
            enabled=(
                default_config.plugins[plugin_name].enabled
                if plugin_name not in existing_config.plugins
                else existing_config.plugins[plugin_name].enabled
            ),
            settings=(
                default_config.plugins[plugin_name].settings
                if plugin_name not in existing_config.plugins
                else existing_config.plugins[plugin_name].settings
            ),
        )

    # Create config with file path for saving
    active_profile = manager.get_active_profile()
    config_file = manager.get_active_profile_directory() / "config.yaml"
    return Config(plugins=plugin_settings, profile=active_profile, file_path=config_file)


async def _select_plugins_to_enable_disable(
    cfg: Config,
) -> None:
    rich.print("")
    rich.print("> 1. Select plugins")
    rich.print("")
    rich.print(
        "> Unpage uses plugins to access your infrastructure resources, like AWS or Aptible."
    )
    rich.print("> Each plugin may be enabled or disabled to match your desired knowledge graph")
    rich.print(
        "> Some plugins require configuration, for example the AWS plugin needs access to your AWS account"
    )
    rich.print("")
    rich.print("> These pre-selected plugins are the one we recommend:")
    rich.print("")
    await _enable_disable_plugins(cfg)
    rich.print("")
    rich.print("> Great! We'll setup these plugins:")
    rich.print(">")
    rich.print(f">   {', '.join(sorted([p for p in cfg.plugins if cfg.plugins[p].enabled]))}")
    rich.print(">")


async def _enable_disable_plugins(cfg: Config) -> None:
    possible_plugins = list(cfg.plugins.keys())
    enabled_search = len(possible_plugins) > 10
    rich.print("")
    selected_plugins = await checkbox(
        "Which plugins would you like to enable? Press enter to continue",
        choices=[Choice(p, checked=True) for p in cfg.plugins],
        use_search_filter=enabled_search,
        use_jk_keys=not enabled_search,
    )
    for plugin_name in cfg.plugins:
        cfg.plugins[plugin_name].enabled = plugin_name in selected_plugins
    rich.print("")


async def _configure_plugins(cfg: Config) -> None:
    rich.print("> 2. Configure plugins")
    rich.print("")
    rich.print("> Now we will go through each plugin and configure it.")
    rich.print(
        "> We will run validation for each plugin, and if something doesn't work you'll have a chance to retry the configuration."
    )
    rich.print(
        "> You can stop the retries, even if the validation is failing, and we'll move to the next plugin."
    )
    rich.print("")
    rich.print("> Ready to start plugin configuration and validation?")
    rich.print("")
    await questionary.press_any_key_to_continue().unsafe_ask_async()
    rich.print("")
    plugin_manager = PluginManager(cfg)
    for plugin_name in cfg.plugins:
        if not cfg.plugins[plugin_name].enabled:
            await _send_event(f"plugin_disabled_{plugin_name}")
            continue
        attempts = 1
        while True:
            rich.print(f"> [bold]{plugin_name}[/bold] plugin configuration:")
            rich.print("")
            cfg.plugins[plugin_name].settings = await plugin_manager.get_plugin(
                plugin_name
            ).interactive_configure()
            plugin_manager = PluginManager(cfg)
            if await _plugin_valid(plugin_manager, plugin_name):
                await _send_event(
                    f"plugin_valid_{plugin_name}", extra_params={"attempts": attempts}
                )
                break
            rich.print(f"> Validation failed for {plugin_name}")
            if not await confirm("Retry?"):
                await _send_event(
                    f"plugin_invalid_{plugin_name}", extra_params={"attempts": attempts}
                )
                break
            attempts += 1
            rich.print("")
        rich.print("")
    rich.print("> Wooooo you did it! All the plugins are configured and ready to use!")
    rich.print("")
    rich.print("> Ready to move on?")
    rich.print("")
    await questionary.press_any_key_to_continue().unsafe_ask_async()
    rich.print("")


async def _plugin_valid(plugin_manager: PluginManager, plugin_name: str) -> bool:
    rich.print(f"> Validating {plugin_name}...")
    try:
        await plugin_manager.get_plugin(plugin_name).validate_plugin_config()
    except Exception as ex:
        rich.print(f"Error validating {plugin_name}:\n{ex}")
        return False
    rich.print(f"[green]{plugin_name} configuration is valid![/green]")
    return True


async def _suggest_building_graph(use_uv_run: bool) -> None:
    rich.print("> 3. Next steps")
    rich.print("")
    rich.print("> Now you're all set to build the infrastructure knowledge graph.")
    rich.print(
        "> The infrastructure knowledge graph is a directed graph of nodes and edges that represent your infra."
    )
    rich.print(
        "> Nodes are resources like ec2 instances, databases, load balancers, and much more."
    )
    rich.print("> Edges represent a directional relationship between two nodes.")
    rich.print("")
    rich.print(
        "> The knowledge graph is a powerful tool for agents to use when investigating incidents!"
    )
    rich.print("")
    rich.print("> Create the graph by running:")
    rich.print(">")
    rich.print(f">   {'uvx' if not use_uv_run else 'uv run'} unpage graph build")
    rich.print(">")
    rich.print("> This is full usage for `unpage graph build`:")
    rich.print("")
    graph_build_cmd = " ".join([a if a != "configure" else "graph build" for a in sys.argv])
    rich.print("> $ unpage graph build --help")
    rich.print("")
    await (await asyncio.create_subprocess_shell(f"{graph_build_cmd} --help")).wait()
    rich.print("")
    rich.print(">")
    if not await confirm(
        "Would you like to launch `unpage graph build` now? (this can take 5min or up to an hour, or even more, depending on how large your infra is)"
    ):
        await _send_event("done_no_graph_build")
        return
    rich.print(">")
    await _send_event("starting_graph_build")
    _replace_current_proc_with_unpage_graph_build(graph_build_cmd)


def _replace_current_proc_with_unpage_graph_build(graph_build_cmd: str) -> None:
    rich.print("> Running: unpage graph build")
    rich.print("")
    cmd = shlex.split(graph_build_cmd)
    os.execvp(cmd[0], cmd)  # noqa: S606 Starting a process without a shell



import sys

import rich

from unpage.cli.profile._app import profile_app
from unpage.config import manager


@profile_app.command
def use(profile_name: str) -> None:
    """Set the active profile.

    Parameters
    ----------
    profile_name
        The name of the profile to use
    """
    # Check if profile exists first
    available_profiles = manager.list_profiles()
    if profile_name not in available_profiles:
        rich.print(f"[red]Profile '{profile_name}' does not exist![/red]")
        rich.print(f"Available profiles: {', '.join(available_profiles)}")
        sys.exit(1)

    manager.set_active_profile(profile_name)
    rich.print(f"[green]Switched to profile '{profile_name}'[/green]")



from unpage.plugins.datadog.nodes.base import DatadogEntityNode


class DatadogSystem(DatadogEntityNode):
    """A node for a Datadog system."""

    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            f"system:default/{self.raw_data['attributes']['name']}",
        ]



import os
from typing import Any

import litellm
import questionary
import rich
from litellm import acompletion

from unpage.config import PluginSettings
from unpage.plugins.base import Plugin
from unpage.utils import Choice, classproperty, select


class LlmPlugin(Plugin):
    """A plugin for configuring LLM models."""

    def __init__(
        self,
        *args: Any,
        model: str,
        api_key: str,
        temperature: float,
        max_tokens: int,
        cache: bool,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.model = model
        self.api_key = api_key
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.cache = cache
        if (
            self.model.startswith("bedrock/")
            and not os.environ.get("AWS_REGION")
            and not os.environ.get("AWS_DEFAULT_REGION")
        ):
            os.environ["AWS_REGION"] = "us-east-1"

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return {
            "model": "openai/gpt-4o-mini",
            "api_key": "",
            "temperature": 0,
            "max_tokens": 8192,
            "cache": True,
        }

    async def interactive_configure(self) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        defaults = self.default_plugin_settings
        recommended_models = {
            "openai": {
                "title": "OpenAI (recommended)",
                "description": "Models from OpenAI https://platform.openai.com/docs/models",
                "models": {
                    "gpt-4o": {
                        "title": "gpt-4o (recommended)",
                        "description": "Fast, intelligent, flexible GPT model https://platform.openai.com/docs/models/gpt-4o",
                    },
                    "gpt-4o-mini": {
                        "description": "Fast, affordable small model for focused tasks https://platform.openai.com/docs/models/gpt-4o-mini",
                    },
                },
            },
            "anthropic": {
                "title": "Anthropic",
                "description": "Models from Anthropic https://www.anthropic.com/claude",
                "models": {
                    "claude-4-sonnet-20250514": {
                        "title": "claude-4-sonnet-20250514 (recommended)",
                        "description": "High intelligence and balanced performance https://docs.anthropic.com/en/docs/about-claude/models/overview#model-comparison-table",
                    },
                    "claude-4-opus-20250514": {
                        "description": "Highest level of intelligence and capability https://docs.anthropic.com/en/docs/about-claude/models/overview#model-comparison-table",
                    },
                },
            },
            "bedrock": {
                "title": "Amazon Bedrock",
                "description": "Models from the Amazon Bedrock marketplace https://aws.amazon.com/bedrock/",
                "models": {
                    "us.anthropic.claude-sonnet-4-20250514-v1:0": {
                        "title": "us.anthropic.claude-sonnet-4-20250514-v1:0 (recommended)",
                        "description": "High intelligence and balanced performance, billed and run through your AWS account",
                    },
                    "us.anthropic.claude-opus-4-20250514-v1:0": {
                        "description": "Highest level of intelligence and capability, billed and run through your AWS account",
                    },
                    "eu.anthropic.claude-sonnet-4-20250514-v1:0": {
                        "description": "High intelligence and balanced performance, billed and run through your AWS account",
                    },
                    "eu.anthropic.claude-opus-4-20250514-v1:0": {
                        "description": "Highest level of intelligence and capability, billed and run through your AWS account",
                    },
                },
            },
        }
        rich.print("Unpage uses LiteLLM and supports all models that LiteLLM supports.")
        rich.print(
            "To configure a model not referenced in this quickstart, you can directly edit Unpage's config.yaml."
        )
        rich.print("More information here: https://docs.unpage.ai/plugins/llm")
        rich.print("")
        provider = await select(
            "Which LLM provider would you like to use?",
            choices=[
                Choice(
                    title=d.get("title", provider),
                    value=provider,
                    description=d["description"],
                )
                for provider, d in recommended_models.items()
            ],
        )
        model = await select(
            f"Which {provider} LLM model would you like to use?",
            choices=[
                Choice(
                    title=d.get("title", model),
                    value=model,
                    description=d["description"],
                )
                for model, d in recommended_models[provider]["models"].items()
            ],
        )
        max_tokens_for_model = litellm.model_cost[model]["max_tokens"]
        return {
            "model": f"{provider}/{model}",
            "api_key": await questionary.password(
                "API key",
                default=self.api_key or defaults["api_key"],
            ).unsafe_ask_async(),
            "temperature": defaults["temperature"],
            "max_tokens": max_tokens_for_model,
            "cache": defaults["cache"],
        }

    async def validate_plugin_config(self) -> None:
        params = {
            "model": self.model,
            "api_key": self.api_key,
            **({"temperature": self.temperature} if not self.model.startswith("bedrock/") else {}),
            "max_tokens": self.max_tokens,
            "cache": self.cache,
        }
        await acompletion(
            **params,
            messages=[
                {
                    "role": "user",
                    "content": "hiiiii",
                }
            ],
        )



from collections.abc import AsyncGenerator, Callable
from typing import Any

import httpx
from pydantic import AwareDatetime, BaseModel


class PapertrailLogEvent(BaseModel):
    generated_at: AwareDatetime
    """time that Papertrail received the log event (ISO 8601 timestamp)

    received_at and display_received_at are in the time zone of the API token owner (set in the profile).
    """
    received_at: AwareDatetime

    """log event message (string)"""
    message: str


class SearchResult(BaseModel):
    """Search result from Papertrail API

    The search endpoint returns a JSON object containing these keys:

        events: an array of hashes of log events (one hash per event)
        min_id: the lowest event ID examined
        max_id: the highest event ID examined

    In addition to min_id and max_id, the response object may also contain min_time_at or max_time_at. These are the oldest or newest timestamps searched during this request. These keys are useful for displaying the time the search spanned in a human-friendly way, but since multiple events may have occured during the same second, they should not be used to select a time range for subsequent queries.

    Each search query will return one of three types of result sets. The client can use the type to decide what to do next. The 3 possible types are:

        * when searching backward: a full set of matching log messages with reached_beginning set to true. All matching events from the set max_id or max_time back to the beginning of searchable retention were examined.
        * when searching forward: a full set of matching log messages with reached_end set to true. All matching events from the set min_id or min_time up to the present were examined.
        * limit-size set of matching log messages (default 1000). The events array contains these elements, and reached_record_limit is included in the result.
        * partial set of matching messages with reached_time_limit set to true. Papertrail's per-request time limit was reached before a full set of events was found.

    Papertrail's workload varies significantly based on the number of possible events and the complexity of the query. To ensure that the caller does not block forever, the search API automatically enforces a per-request time limit of about 5 seconds. When the response contains reached_time_limit, this time limit expired before the full set of results was returned.
    """

    """An array of hashes of log events (one hash per event)"""
    events: list[PapertrailLogEvent]
    """the lowest event ID examined"""
    min_id: str
    """the highest event ID examined"""
    max_id: str
    """In addition to min_id and max_id, the response object may also contain min_time_at or max_time_at. These are the oldest or newest timestamps searched during this request. These keys are useful for displaying the time the search spanned in a human-friendly way, but since multiple events may have occured during the same second, they should not be used to select a time range for subsequent queries."""
    min_time_at: AwareDatetime | None = None
    """In addition to min_id and max_id, the response object may also contain min_time_at or max_time_at. These are the oldest or newest timestamps searched during this request. These keys are useful for displaying the time the search spanned in a human-friendly way, but since multiple events may have occured during the same second, they should not be used to select a time range for subsequent queries."""
    max_time_at: AwareDatetime | None = None
    """When searching backward: a full set of matching log messages with reached_beginning set to true. All matching events from the set max_id or max_time back to the beginning of searchable retention were examined."""
    reached_beginning: bool = False
    """When searching forward: a full set of matching log messages with reached_end set to true. All matching events from the set min_id or min_time up to the present were examined."""
    reached_end: bool = False
    """A limit-size set of matching log messages (default 1000). The events array contains these elements, and reached_record_limit is included in the result."""
    reached_record_limit: bool = False
    """A partial set of matching messages with reached_time_limit set to true. Papertrail's per-request time limit was reached before a full set of events was found.

    Papertrail's workload varies significantly based on the number of possible events and the complexity of the query. To ensure that the caller does not block forever, the search API automatically enforces a per-request time limit of about 5 seconds. When the response contains reached_time_limit, this time limit expired before the full set of results was returned.
    """
    reached_time_limit: bool = False


class PapertrailClient(httpx.AsyncClient):
    """Papertrail API client

    https://www.papertrail.com/help/search-api/
    """

    BASE_URL = "https://papertrailapp.com/api/v1"

    def __init__(self, token: str, connection_limit: int = 10, base_url: str = BASE_URL) -> None:
        self._token = token
        super().__init__(
            base_url=base_url,
            headers={
                "X-Papertrail-Token": self._token,
            },
            timeout=httpx.Timeout(10),
            limits=httpx.Limits(max_connections=connection_limit),
        )

    async def verify_connection(self) -> None:
        response = await self.get("/events/search.json", params={"limit": 1})
        response.raise_for_status()
        SearchResult.model_validate(response.json())

    async def search(
        self,
        query: str,
        min_time: AwareDatetime,
        max_time: AwareDatetime,
        min_time_at_callback: Callable[[AwareDatetime], None] | None = None,
        continue_search: Callable[[AwareDatetime | None], bool] | None = None,
    ) -> AsyncGenerator[PapertrailLogEvent, None]:
        max_id = None
        logs_found = 0
        while True:
            params: dict[str, Any] = {
                "q": query,
                "tail": False,
                **(
                    {
                        "min_time": int(min_time.timestamp()),
                        "max_time": int(max_time.timestamp()),
                        "limit": 100,
                    }
                    if max_id is None
                    else {
                        "max_id": max_id,
                        "limit": 1000,
                    }
                ),
            }
            response = await self.get("/events/search.json", params=params)
            response.raise_for_status()
            result = SearchResult.model_validate(response.json())
            event_before_min_time = False
            for event in result.events:
                if event.generated_at >= min_time or event.received_at >= min_time:
                    yield event
                else:
                    event_before_min_time = True
            max_id = result.min_id
            if event_before_min_time:
                break
            logs_found += len(result.events)
            if result.reached_end:
                break
            if result.reached_beginning:
                break
            if result.min_time_at is not None and result.min_time_at < min_time:
                break
            if min_time_at_callback is not None and result.min_time_at is not None:
                min_time_at_callback(result.min_time_at)
            if continue_search is not None and not continue_search(result.min_time_at):
                break



"""This module is a custom build backend that wraps `hatching.build`.

It patches the `pyproject.toml` file with pinned dependencies from `uv.lock`
before building the wheel or sdist.
"""

import functools
import shutil
import subprocess
from collections.abc import Callable
from pathlib import Path
from typing import Any, cast

import tomlkit
from hatchling import build

PROJECT_ROOT = Path(__file__).parent


def _get_pinned_dependencies() -> list[str]:
    """Get pinned dependencies from uv.lock"""
    uv = shutil.which("uv")

    if uv is None:
        raise FileNotFoundError("uv command not found. Please install uv first.")

    result = subprocess.run(
        [
            uv,
            "export",
            "--format=requirements-txt",
            "--no-header",
            "--no-hashes",
            "--no-editable",
            "--no-dev",
            "--no-annotate",
            "--no-emit-project",
        ],
        cwd=PROJECT_ROOT,
        capture_output=True,
        text=True,
        check=True,
    )

    return result.stdout.strip().splitlines()


def _with_pinned_dependencies(func: Callable) -> Callable:
    """Decorator that temporarily patches pyproject.toml with pinned dependencies"""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        pyproject_path = Path("pyproject.toml")
        backup_path = Path("pyproject.toml.backup")

        # Read current config
        config = tomlkit.loads(pyproject_path.read_text())

        # Backup original
        shutil.copy2(pyproject_path, backup_path)

        try:
            # Update with pinned dependencies
            project = cast("dict[str, Any]", config["project"])
            project["dependencies"] = _get_pinned_dependencies()
            config["project"] = project

            # Write modified version
            pyproject_path.write_text(tomlkit.dumps(config))

            # Call the original function
            return func(*args, **kwargs)

        finally:
            # Restore original pyproject.toml
            if backup_path.exists():
                shutil.move(backup_path, "pyproject.toml")

    return wrapper


# Functions that trigger actual builds and need dependency pinning
BUILD_FUNCTIONS = {"build_wheel", "build_sdist"}


def __getattr__(name: str) -> Any:  # noqa: ANN401
    """Dynamically proxy all attributes from `hatchling.build`.

    Wraps build functions with dependency pinning and passes through all other
    calls unchanged.
    """
    if not hasattr(build, name):
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

    attr = getattr(build, name)

    # If it's a build function, wrap it with dependency pinning
    if name in BUILD_FUNCTIONS and callable(attr):
        return _with_pinned_dependencies(attr)

    # Otherwise, pass through unchanged
    return attr



import asyncio
import logging
import sys
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from aioboto3 import Session
from botocore.exceptions import ClientError, SSOTokenLoadError, TokenRetrievalError


async def list_accessible_regions_for_service(session: Session, service_name: str) -> list[str]:
    """Return a list of regions that the current credentials can access."""

    async def _check_region(region: str) -> tuple[str, bool]:
        """Attempt to call an API in the given region. Return True if accessible."""
        try:
            async with session.client("sts", region_name=region) as client:
                await client.get_caller_identity()
            return region, True
        except ClientError:
            return region, False

    # Now, check access to each region (concurrently)
    results = await asyncio.gather(
        *(_check_region(region) for region in await session.get_available_regions(service_name))
    )

    return [region for region, success in results if success]


@asynccontextmanager
async def swallow_boto_client_access_errors(
    service_name: str | None = None, region: str | None = None
) -> AsyncIterator[None]:
    try:
        yield
    except ClientError as e:
        err = e.response.get("Error", {})
        error_code = err.get("Code", None)
        error_msg = err.get("Message", "Unknown Message")
        if error_code in [
            "AccessDenied",
            "AccessDeniedException",
            "AuthorizationError",
            "UnauthorizedOperation",
        ]:
            print(
                f"Ignoring access denied ({error_code}) for {f'{service_name}.' if service_name else ''}{e.operation_name}{f' in {region}' if region else ''}: {error_msg}",
                file=sys.stderr,
            )
            return
        raise


@asynccontextmanager
async def hide_traceback_for_failed_sso_logins() -> AsyncIterator[None]:
    aiologger = logging.getLogger("aiobotocore")
    original_aiologger_level = aiologger.level
    try:
        aiologger.setLevel(logging.ERROR)
        yield
    finally:
        aiologger.setLevel(original_aiologger_level)


async def run_sso_login(profile: str | None = None) -> None:
    """Run the AWS SSO login command for the configured profile."""
    try:
        print(f"Attempting to login using AWS SSO for profile '{profile}'")
        print(f"Running 'aws sso login --profile {profile}'")
        print("Please complete the SSO authentication in your browser...")

        cmd = ["aws", "sso", "login"]
        if profile:
            cmd.extend(["--profile", profile])

        process = await asyncio.create_subprocess_exec(
            *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )
        stdout, stderr = await process.communicate()

        if process.returncode != 0:
            error_msg = stderr.decode() if stderr else "Unknown error"
            print(f"Error during SSO login: {error_msg}")
            raise RuntimeError(f"Failed to authenticate with AWS SSO: {error_msg}")

    except Exception as e:
        print(f"Unexpected error during SSO login: {e}")
        raise
    else:
        print("SSO login successful")


async def ensure_aws_session(session: Session) -> None:
    async with hide_traceback_for_failed_sso_logins(), session.client("sts") as client:
        try:
            await client.get_caller_identity()
        except (SSOTokenLoadError, TokenRetrievalError):
            await run_sso_login(session.profile_name)
            await client.get_caller_identity()
        except Exception as e:
            print(f"Error using session to call STS GetCallerIdentity: {e}")
            raise
